@@ -1,6 +1,6 @@
 """
-SQLObject 0.9.3
----------------
+SQLObject 0.10.4
+----------------
 
 :author: Ian Bicking <ianb@colorstudy.com>
 
@@ -41,18 +41,24 @@ import declarative
 import events
 from sresults import SelectResults
 from formencode import schema, compound
+from util.threadinglocal import local
 
 import sys
 if sys.version_info[:3] < (2, 2, 0):
     raise ImportError, "SQLObject requires Python 2.2.0 or later"
 
+"""
+This thread-local storage is needed for RowCreatedSignals. It gathers
+code-blocks to execute _after_ the whole hierachy of inherited SQLObjects
+is created. See SQLObject._create
+"""
+_postponed_local = local()
+
 NoDefault = sqlbuilder.NoDefault
 
 class SQLObjectNotFound(LookupError): pass
 class SQLObjectIntegrityError(Exception): pass
 
-True, False = 1==1, 0==1
-
 def makeProperties(obj):
     """
     This function takes a dictionary of methods and finds
@@ -90,8 +96,7 @@ def makeProperties(obj):
         if len(setters) == 1 and setters.has_key('doc'):
             continue
         if d.has_key(var):
-            if isinstance(d[var], types.MethodType) \
-                   or isinstance(d[var], types.FunctionType):
+            if isinstance(d[var], (types.MethodType, types.FunctionType)):
                 warnings.warn(
                     "I tried to set the property %r, but it was "
                     "already set, as a method (%r).  Methods have "
@@ -489,6 +494,7 @@ class sqlmeta(object):
 
         if soClass._SO_finishedClassCreation:
             unmakeProperties(soClass)
+            makeProperties(soClass)
 
         for func in post_funcs:
             func(soClass, column)
@@ -583,6 +589,7 @@ class sqlmeta(object):
 
         if soClass._SO_finishedClassCreation:
             unmakeProperties(soClass)
+            makeProperties(soClass)
 
     delJoin = classmethod(delJoin)
 
@@ -601,12 +608,15 @@ class sqlmeta(object):
     ## Utility methods
     ########################################
 
+    def getColumns(sqlmeta):
+        return sqlmeta.columns.copy()
+
     def asDict(self):
         """
         Return the object as a dictionary of columns to values.
         """
         result = {}
-        for key in self.columns:
+        for key in self.getColumns():
             result[key] = getattr(self.instance, key)
         result['id'] = self.instance.id
         return result
@@ -657,8 +667,8 @@ def deprecated(message, level=1, stacklevel=2):
     if warnings_level is not None and warnings_level <= level:
         warnings.warn(message, DeprecationWarning, stacklevel=stacklevel)
 
-if sys.version_info[:3] < (2, 3, 0):
-    deprecated("Support for Python 2.2 has been declared obsolete and will be removed in the next release of SQLObject")
+#if sys.version_info[:3] < (2, 4, 0):
+#    deprecated("Support for Python 2.3 has been declared obsolete and will be removed in the next release of SQLObject")
 
 def setDeprecationLevel(warning=1, exception=None):
     """
@@ -711,9 +721,6 @@ class SQLObject(object):
     _inheritable = False # Is this class inheritable?
     _parent = None # A reference to the parent instance
     childName = None # Children name (to be able to get a subclass)
-    # moved to sqlmeta in 0.8:
-    _parentClass = _sqlmeta_attr('parentClass', 1)
-    _childClasses = _sqlmeta_attr('childClasses', 1)
 
     # The law of Demeter: the class should not call another classes by name
     SelectResultsClass = SelectResults
@@ -802,21 +809,7 @@ class SQLObject(object):
 
         classregistry.registry(cls.sqlmeta.registry).addClass(cls)
 
-    _style = _sqlmeta_attr('style', 1)
-    _table = _sqlmeta_attr('table', 1)
-    _idName = _sqlmeta_attr('idName', 1)
-    _lazyUpdate = _sqlmeta_attr('lazyUpdate', 1)
-    _defaultOrder = _sqlmeta_attr('defaultOrder', 1)
-    _cacheValues = _sqlmeta_attr('cacheValues', 1)
-    _registry = _sqlmeta_attr('registry', 1)
-    _idType = _sqlmeta_attr('idType', 1)
-    _fromDatabase = _sqlmeta_attr('fromDatabase', 1)
-    _expired = _sqlmeta_attr('expired', 2)
-    addColumn = _sqlmeta_attr('addColumn', 1)
-    delColumn = _sqlmeta_attr('delColumn', 1)
-    addJoin = _sqlmeta_attr('addJoin', 1)
-    delJoin = _sqlmeta_attr('delJoin', 1)
-    addIndex = _sqlmeta_attr('addIndex', 1)
+    _expired = _sqlmeta_attr('expired', 1)
 
     # @classmethod
     def _SO_setupSqlmeta(cls, new_attrs, is_base):
@@ -830,8 +823,7 @@ class SQLObject(object):
         if (not new_attrs.has_key('sqlmeta')
             and not is_base):
             # We have to create our own subclass, usually.
-            # type(className, bases_tuple, attr_dict) creates a new
-            # subclass:
+            # type(className, bases_tuple, attr_dict) creates a new subclass.
             cls.sqlmeta = type('sqlmeta', (cls.sqlmeta,), {})
         if not issubclass(cls.sqlmeta, sqlmeta):
             # We allow no superclass and an object superclass, instead
@@ -869,22 +861,11 @@ class SQLObject(object):
         been deprecated; they are moved to the sqlmeta class, and
         a deprecation warning is given.
         """
-        for attr in ['_table', '_lazyUpdate', '_style', '_idName',
-                     '_defaultOrder', '_cacheValues', '_registry',
-                     '_idType', '_fromDatabase', '_idSequence']:
-            if new_attrs.has_key(attr):
-                new_name = attr[1:]
-                deprecated("%r is deprecated; please set the %r "
-                           "attribute in sqlmeta instead" %
-                           (attr, new_name), level=1,
-                           stacklevel=5)
-                setattr(cls.sqlmeta, new_name, new_attrs[attr])
-                delattr(cls, attr)
         for attr in ['_expired']:
             if new_attrs.has_key(attr):
                 deprecated("%r is deprecated and read-only; please do "
                            "not use it in your classes until it is fully "
-                           "deprecated" % attr, level=2, stacklevel=5)
+                           "deprecated" % attr, level=1, stacklevel=5)
 
     _SO_cleanDeprecatedAttrs = classmethod(_SO_cleanDeprecatedAttrs)
 
@@ -1061,8 +1042,8 @@ class SQLObject(object):
         if self.sqlmeta.cacheValues:
             setattr(self, instanceName(name), value)
 
-    def set(self, **kw):
-        if not self.sqlmeta._creating and not getattr(self.sqlmeta, "row_update_sig_suppress", False):
+    def set(self, _suppress_set_sig=False, **kw):
+        if not self.sqlmeta._creating and not getattr(self.sqlmeta, "row_update_sig_suppress", False) and not _suppress_set_sig:
             self.sqlmeta.send(events.RowUpdateSignal, self, kw)
         # set() is used to update multiple values at once,
         # potentially with one SQL statement if possible.
@@ -1180,36 +1161,59 @@ class SQLObject(object):
             return joinClass.get(id)
 
     def __init__(self, **kw):
-        # We shadow the sqlmeta class with an instance of sqlmeta
-        # that points to us (our sqlmeta buddy object; where the
-        # sqlmeta class is our class's buddy class)
-        self.sqlmeta = self.__class__.sqlmeta(self)
-        # The get() classmethod/constructor uses a magic keyword
-        # argument when it wants an empty object, fetched from the
-        # database.  So we have nothing more to do in that case:
-        if kw.has_key('_SO_fetch_no_create'):
-            return
+        # If we are the outmost constructor of a hiearchy of
+        # InheritableSQLObjects (or simlpy _the_ constructor of a "normal"
+        # SQLObject), we create a threadlocal list that collects the
+        # RowCreatedSignals, and executes them if this very constructor is left
+        try:
+            _postponed_local.postponed_calls
+            postponed_created = False
+        except AttributeError:
+            _postponed_local.postponed_calls = []
+            postponed_created = True
 
-        post_funcs = []
-        self.sqlmeta.send(events.RowCreateSignal, kw, post_funcs)
+        try:
+            # We shadow the sqlmeta class with an instance of sqlmeta
+            # that points to us (our sqlmeta buddy object; where the
+            # sqlmeta class is our class's buddy class)
+            self.sqlmeta = self.__class__.sqlmeta(self)
+            # The get() classmethod/constructor uses a magic keyword
+            # argument when it wants an empty object, fetched from the
+            # database.  So we have nothing more to do in that case:
+            if kw.has_key('_SO_fetch_no_create'):
+                return
 
-        # Pass the connection object along if we were given one.
-        if kw.has_key('connection'):
-            self._connection = kw['connection']
-            self.sqlmeta._perConnection = True
-            del kw['connection']
+            post_funcs = []
+            self.sqlmeta.send(events.RowCreateSignal, kw, post_funcs)
 
-        self._SO_writeLock = threading.Lock()
+            # Pass the connection object along if we were given one.
+            if kw.has_key('connection'):
+                self._connection = kw['connection']
+                self.sqlmeta._perConnection = True
+                del kw['connection']
 
-        if kw.has_key('id'):
-            id = self.sqlmeta.idType(kw['id'])
-            del kw['id']
-        else:
-            id = None
+            self._SO_writeLock = threading.Lock()
 
-        self._create(id, **kw)
-        for func in post_funcs:
-            func(self)
+            if kw.has_key('id'):
+                id = self.sqlmeta.idType(kw['id'])
+                del kw['id']
+            else:
+                id = None
+
+            self._create(id, **kw)
+
+            for func in post_funcs:
+                func(self)
+        finally:
+            # if we are the creator of the tl-storage, we
+            # have to exectute and under all circumstances
+            # remove the tl-storage
+            if postponed_created:
+                try:
+                    for func in _postponed_local.postponed_calls:
+                        func()
+                finally:
+                    del _postponed_local.postponed_calls
 
     def _create(self, id, **kw):
 
@@ -1228,8 +1232,14 @@ class SQLObject(object):
 
                 # If we don't get it, it's an error:
                 # If we specified an SQL DEFAULT, then we should use that
-                if default is NoDefault and column.defaultSQL is not None:
-                    raise TypeError, "%s() did not get expected keyword argument %s" % (self.__class__.__name__, column.name)
+                if default is NoDefault:
+                    if column.defaultSQL is None:
+                        raise TypeError, "%s() did not get expected keyword argument '%s'" % (self.__class__.__name__, column.name)
+                    else:
+                        # There is defaultSQL for the column - do not put
+                        # the column to kw so that the backend creates the value
+                        continue
+
                 # Otherwise we put it in as though they did pass
                 # that keyword:
 
@@ -1268,9 +1278,11 @@ class SQLObject(object):
         self._init(id)
         post_funcs = []
         kw = dict([('class', self.__class__), ('id', id)])
-        self.sqlmeta.send(events.RowCreatedSignal, kw, post_funcs)
-        for func in post_funcs:
-            func(self)
+        def _send_RowCreatedSignal():
+            self.sqlmeta.send(events.RowCreatedSignal, kw, post_funcs)
+            for func in post_funcs:
+                func(self)
+        _postponed_local.postponed_calls.append(_send_RowCreatedSignal)
 
     def _SO_getID(self, obj):
         return getID(obj)
@@ -1280,12 +1292,17 @@ class SQLObject(object):
             column = cls.sqlmeta.columns[name]
             if isinstance(column, col.SOUnicodeCol):
                 value = value.encode(column.dbEncoding)
+        if isinstance(name, str):
+            name = (name,)
+            value = (value,)
+        if len(name) != len(value):
+            raise ValueError, "'column' and 'value' tuples must be of the same size"
+        condition = sqlbuilder.AND(*[getattr(cls.q, n)==v for n,v in zip(name, value)])
         return (connection or cls._connection)._SO_selectOneAlt(
             cls,
             [cls.sqlmeta.idName] +
             [column.dbName for column in cls.sqlmeta.columnList],
-            dbName,
-            value), None
+            condition), None
     _findAlternateID = classmethod(_findAlternateID)
 
     def _SO_fetchAlternateID(cls, name, dbName, value, connection=None, idxName=None):
@@ -1410,9 +1427,6 @@ class SQLObject(object):
     def createJoinTables(cls, ifNotExists=False, connection=None):
         conn = connection or cls._connection
         for join in cls._getJoinsToCreate():
-            if not getattr(join, 'createRelatedTable', True):
-                # This join has requested not to be created
-                continue
             if (ifNotExists and
                 conn.tableExists(join.intermediateTable)):
                 continue
@@ -1423,9 +1437,6 @@ class SQLObject(object):
         conn = connection or cls._connection
         sql = []
         for join in cls._getJoinsToCreate():
-            if not getattr(join, 'createRelatedTable', True):
-                # This join has requested not to be created
-                continue
             sql.append(conn._SO_createJoinTableSQL(join))
         return ';\n'.join(sql)
     createJoinTablesSQL = classmethod(createJoinTablesSQL)
@@ -1453,7 +1464,7 @@ class SQLObject(object):
         for join in cls.sqlmeta.joins:
             if not join:
                 continue
-            if not join.hasIntermediateTable():
+            if not join.hasIntermediateTable() or not getattr(join, 'createRelatedTable', True):
                 continue
             if join.soClass.__name__ > join.otherClass.__name__:
                 continue
@@ -1466,7 +1477,7 @@ class SQLObject(object):
         for join in cls.sqlmeta.joins:
             if not join:
                 continue
-            if not join.hasIntermediateTable():
+            if not join.hasIntermediateTable() or not getattr(join, 'createRelatedTable', True):
                 continue
             if join.soClass.__name__ > join.otherClass.__name__:
                 continue
@@ -1488,7 +1499,8 @@ class SQLObject(object):
     clearTable = classmethod(clearTable)
 
     def destroySelf(self):
-        self.sqlmeta.send(events.RowDestroySignal, self)
+        post_funcs = []
+        self.sqlmeta.send(events.RowDestroySignal, self, post_funcs)
         # Kills this object.  Kills it dead!
 
         klass = self.__class__
@@ -1550,6 +1562,9 @@ class SQLObject(object):
         self._connection._SO_delete(self)
         self._connection.cache.expire(self.id, self.__class__)
 
+        for func in post_funcs:
+            func()
+
     def delete(cls, id, connection=None):
         obj = cls.get(id, connection=connection)
         obj.destroySelf()
@@ -1578,6 +1593,9 @@ class SQLObject(object):
                   self.id,
                   ' '.join(['%s=%s' % (name, repr(value)) for name, value in self._reprItems()]))
 
+    def __sqlrepr__(self, db):
+        return str(self.id)
+
     def sqlrepr(cls, value, connection=None):
         return (connection or cls._connection).sqlrepr(value)
 
@@ -1602,11 +1620,36 @@ class SQLObject(object):
         return items
 
     def setConnection(cls, value):
-        if isinstance(value, (str, unicode)):
+        if isinstance(value, basestring):
             value = dbconnection.connectionForURI(value)
         cls._connection = value
+
     setConnection = classmethod(setConnection)
 
+    def tablesUsedImmediate(self):
+        return [self.__class__.q]
+
+    def __eq__(self, other):
+        if self.__class__ is other.__class__:
+            if self.id == other.id:
+                return True
+        return False
+
+    def __ne__(self, other):
+        return not self.__eq__(other)
+
+    def __lt__(self, other):
+       return NotImplemented
+
+    def __le__(self, other):
+       return NotImplemented
+
+    def __gt__(self, other):
+       return NotImplemented
+
+    def __ge__(self, other):
+       return NotImplemented
+
 
 def capitalize(name):
     return name[0].capitalize() + name[1:]
@@ -1637,11 +1680,11 @@ class SQLObjectState(object):
 def getID(obj):
     if isinstance(obj, SQLObject):
         return obj.id
-    elif type(obj) is type(1):
+    elif isinstance(obj, int):
         return obj
-    elif type(obj) is type(1L):
+    elif isinstance(obj, long):
         return int(obj)
-    elif type(obj) is type(""):
+    elif isinstance(obj, str):
         try:
             return int(obj)
         except ValueError:
@@ -1650,11 +1693,11 @@ def getID(obj):
         return None
 
 def getObject(obj, klass):
-    if type(obj) is type(1):
+    if isinstance(obj, int):
         return klass(obj)
-    elif type(obj) is type(1L):
+    elif isinstance(obj, long):
         return klass(int(obj))
-    elif type(obj) is type(""):
+    elif isinstance(obj, str):
         return klass(int(obj))
     elif obj is None:
         return None
